//=========================================================
//  ToSVCD
//    transcode DVB-VDR recordings to SVCD
//
//  $Id: tosvcd.c 1.29 Mon, 09 Dec 2002 10:24:57 +0100 ws $
//
//  (C) Copyright 2002 Werner Schweer (ws@seh.de)
//=========================================================

#define _FILE_OFFSET_BITS 64
#define _LARGEFILE_SOURCE 1
#define _LARGEFILE64_SOURCE 1

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include <time.h>
#include <pthread.h>
#include <list>
#include <values.h>
#include <signal.h>
#include <wait.h>
#include <glob.h>

#include "mm_accel.h"
#include "decode.h"
#include "tosvcd.h"

// program options:
bool verbose    = false;
bool mpegonly   = false;
char* oName     = "mops";     // default output name
int quality     = 4;          // default best quality
int cdsize      = (800-25);   // war 800-30
bool monochrome = false;
int widescreen  = 0;
bool denoise    = false;
bool denoise2   = false;
bool vcdonly    = false;
bool whole      = false;       // transcode whole uncutted movie
int numcds      = -1;          // number of output cd's
char* logfile   = 0;
bool keepFiles  = false;
bool wideScreen = false;

// global state:
enum State { INIT, SYNC, ENCODE, FLUSH, FLUSH1, RESTART };
enum State state;
int doubledPictures;
int droppedPictures;
int totalPictures;
time_t startTime;

// some more configuration values
int restartMinSec = 60;       // first check size after 1 min
int restartSec    = 25*60;    // restart if ouput exceeds target size for this time
int restartAmount = 40;       // restart immediate if 40% too big
int minutesPerCD  = 62;
bool autorestart  = true;
bool transcodeAudio = false;  // transcode audio to 44100 sample rate

MpegDec* mpeg2dec;
int picOffset;
int initPicOffset;
int absPic;                   // current picture of uncutted stream
int curPic;                   // current picture of cutted stream
int videoWritten;
int tooBigStart;

const int CHUNK_MAX = 1024 * 224*2;
unsigned char chunk_buffer[CHUNK_MAX];
unsigned char* chunk_ptr;
unsigned char code;
unsigned int shift;
bool is_sequence_needed;
bool in_slice;

FILE* vPipe;   // video pipe
bool picture_header_found = false;
pthread_t videothread;

typedef std::list<int> MarkList;
typedef MarkList::iterator iMark;
MarkList markList;
int endPicture;

Index* picIndex;
AudioStream audio;
MFile f;
FILE* logf = 0;

int twidth = 480;
int theight = 576;
char pipeCmd[512];
char* pipename;

uint64_t apts;    // audio presentation time stamp as read from input stream
uint64_t vpts;    // video pts
uint64_t vptse;   // video pts
uint64_t adpts;   // audio device output state
uint64_t vdpts;   // video device output state
int64_t  async;

//---------------------------------------------------------
//   log
//---------------------------------------------------------

static void log(bool se, const char* format, ...)
      {
      va_list args;
      va_start(args, format);
      if (logf) {
            vfprintf(logf, format, args);
            fflush(logf);
            }
      if (se)
            vfprintf(stderr, format, args);
      va_end(args);
      }

//---------------------------------------------------------
//   close
//---------------------------------------------------------

void MFile::close()
      {
      if (in) {
            fclose(in);
            in = 0;
            }
      }

//---------------------------------------------------------
//   open
//---------------------------------------------------------

bool MFile::open()
      {
      if (curFile == files)
            return true;
      close();

      log(0, "open file <%s>\n", filenames[curFile]);
      in = fopen(filenames[curFile], "r");
      if (in == 0) {
            log(1, "open inputfile %s failed: %s\n",
                     filenames[curFile], strerror(errno));
            return true;
            }
      ++curFile;
      return false;
      }

//---------------------------------------------------------
//   seekPic
//    return true on error
//---------------------------------------------------------

bool MFile::seekPic(int pic)
      {
      Index* index = &picIndex[pic];
      int pos = index->offset;
      if (index->type != I_FRAME) {
            log(1, "seek to non I-Frame: pic %d offset 0x%x\n", pic, pos);
            return true;
            }
      if (curFile-1 != index->number) {
            curFile = (index->number)-1;
            log(0, "open curFile %d\n", curFile);
            open();
            }

      //---------------------------------------------------
      //    we need the previous PES video header:
      //       search backwards to (00 00 01 e0) header
      //---------------------------------------------------

      for (;;) {
            int rv;
            if (pos < 0)
                  rv = fseek(in, -pos, SEEK_END);
            else
                  rv = fseek(in, pos, SEEK_SET);
            if (rv == -1) {
                  log(1, "\nseek to %d int <%s> failed\n",
                           pos, filenames[curFile-1]);
                  exit(-1);
                  }
            unsigned int a;
            if (fread(&a, 1, 4, in) != 4) {
                  log(1, "\nseek: read error <%s> %s\n", filenames[curFile-1], strerror(errno));
                  exit(-1);
                  }
            if ((a & 0xe0ffffff) == 0xe0010000)
                  break;
            --pos;
            if (pos == -1) {
                  curFile -= 2;
                  if (curFile <= 0)
                        break;
                  open();
                  }
            }
      fseek(in, pos, SEEK_SET);
      return false;
      }

//---------------------------------------------------------
//   get
//---------------------------------------------------------

int MFile::get()
      {
      int c;
      while ((c = getc(in)) == EOF) {
            if (open())
                  break;
            }
      return c;
      }

//---------------------------------------------------------
//   draw
//---------------------------------------------------------

void MpegDec::draw(Frame* f)
      {
      if ((state != ENCODE) && (state != SYNC))
            return;

      if ((curPic > 100) && ((curPic % 25) == 0)) {
            //
            // finish time estimation
            // (TODO: check overflow)
            //
            time_t curTime   = time(0);
            time_t elapsed   = curTime - startTime;
            time_t estimated = totalPictures * (elapsed/10) / (curPic/10) + startTime;
            struct tm* time  = localtime(&estimated);

            //
            // finish size estimation
            //
            long long curSec = curPic/25;
            int percent;
            int framesSec;
            if (elapsed == 0)
                  elapsed = 1;
            if (curSec) {
                  // cursize = Kbyte
                  long long cursize = (videoWritten/1024) + curSec * (audio.bitrate/8);
                  long long tsize   = cursize * (long long)(totalPictures/25) / curSec;
                  percent           = (tsize * 100LL)/ (cdsize*1024);
                  framesSec = curPic*100 / elapsed;
                  }
            else {
                  framesSec = 0;
                  percent = 0;
                  }

            fprintf(stderr, "%3d:%02d %3d:%02d  %2d.%02d/sec finished at %02d.%02d  q: %d  size: %3d%%\r",
               curPic/(25*60), (curPic % (25*60))/25,
               totalPictures/(25*60), (totalPictures % (25*60))/25,
               framesSec/100, framesSec%100,
               time->tm_hour, time->tm_min,
               quality,
               percent
               );
            fflush(stderr);
            log(0, "%3d:%02d finished at %02d.%02d  q: %d  size: %3d%% async %lld\n",
               curPic/(25*60), (curPic % (25*60))/25,
               time->tm_hour, time->tm_min,
               quality,
               percent, async
               );
            int tooBig = percent - 100*numcds;
            if (autorestart && (curSec > restartMinSec) && (tooBig > 0)) {
                  if ((tooBig > restartAmount) || (curSec-tooBigStart > restartSec)) {
                        quality++;
                        state = RESTART;
                        return;
                        }
                  }
            else
                  tooBigStart = curSec;
            }

      uint64_t pts = f->pts;
      if (pts == 0 || apts == 0) {
            log(1, "no pts %lld apts %lld vpts %lld\n", pts, apts, vpts);
            return;
            }
      int n = 0;
      if (state == SYNC) {
            int64_t a = pts - apts - vdpts + adpts;
            n = a / 3600;
            if (n < 0) {
                  ++n;
                  if (n == 0)
                        state = ENCODE;
                  n = 0;  // dont double frame
                  }
            else
                  state = ENCODE;
            log(1, "sync offset %d frames %lld %lld %lld %lld\n", n, pts, apts, vdpts, adpts);
            }
      else {
            if (vptse) {
                  if (vptse != pts) {
                        int64_t vs = pts - vptse;
                        async -= vs;
                        int64_t vsa = vs > 0 ? vs : -vs;
                        if (vsa > 10)
                              log(1, "video stream break %lld = %lld - %lld (%d pics)\n",
                                 vs, pts, vptse, int(vs/3600));

                        }
                  }
            vptse = pts + 3600;
            }
      int picAsync = async / 3600;
      if (picAsync > 0 && picAsync < 20) {  // some sanity checks
            async -= 3600;
            // drop picture
            log(1, "resync: drop pic async now:%lld\n", async);
            n = 0;
            return;
            }
      if (picAsync < 0 && picAsync > -20) {
            n = -picAsync;
            async += 3600 * n;
            log(1, "resync: dup pic %d, async now:%lld\n", n, async);
            }
      do {
            ++absPic;
            ++curPic;
            int size = width() * height();
            fprintf(vPipe, "FRAME\n");
            fwrite(f->base[0], 1, size,   vPipe);
            fwrite(f->base[1], 1, size/4, vPipe);
            fwrite(f->base[2], 1, size/4, vPipe);
            vdpts += 3600;
            } while (n--);

      if (ferror(vPipe)) {
            log(1, "\nwrite of YUV output failed: %s\n", strerror(errno));
            exit(1);
            }
      if (absPic >= endPicture) {
            log(0, "END: absPic %d > endPic %d\n", absPic, endPicture);
            state = FLUSH;
            }
      }

//---------------------------------------------------------
//   nextStartCode
//---------------------------------------------------------

static bool nextStartCode()
      {
      int code = 0;
      int n    = 0;
      int c;
      while ((c = f.get()) != EOF) {
            ++n;
            code <<= 8;
            code |= (c & 0xff);
            if ((code & 0xffffff) == 0x000001) {
                  if (n > 3) {
                        log(1, "\nnextStartCode: skip %d\n", n - 3);
                        // return false;
                        }
                  return true;
                  }
            else if (n > 3)
                  log(0, "%02x ", c);
            }
      return false;
      }

//---------------------------------------------------------
//   parseChunk
//    parse video chunk
//---------------------------------------------------------

static void parseChunk()
      {
      if (is_sequence_needed && (code != 0xb3))
            return;

      bool is_frame_done = in_slice && (!code || (code >= 0xb0));

      if (is_frame_done) {
      	in_slice = false;
            mpeg2dec->frameDone();
            }

      switch (code) {
            case 0x00:	/* picture_start_code */
                  mpeg2dec->processPictureHeader(chunk_buffer);
                  picture_header_found = true;
                  break;

            case 0xb3:	/* sequence_header_code */
                  if (mpeg2dec->processSequenceHeader(chunk_buffer)) {
                        log(0, "bad sequence header\n");
                        exit(1);
                        }
                  if (is_sequence_needed) {
                        is_sequence_needed = false;
                        mpeg2dec->startSequence();
                        }
                  break;

            case 0xb5:
                  if (mpeg2dec->processExtension(chunk_buffer)) {
                        log(0, "\nbad extension\n");
                        exit(1);
                        }
                  break;

            default:
                  if (code >= 0xb0)
                        break;

                  if (!in_slice) {
                        in_slice = true;
                        mpeg2dec->startSlice();
                        }
                  mpeg2dec->processSlice(code, chunk_buffer);
                  break;
            }
      return;
      }

//---------------------------------------------------------
//   videoSequence
//---------------------------------------------------------

static void videoSequence(int end)
      {
      int current = 0;

      while (current != end) {
            unsigned char byte;
            for (;;) {
                  byte = f.get();
                  ++current;
                  if (shift == 0x00000100)
                        break;
                  *chunk_ptr++ = byte;
                  if (chunk_ptr >= (chunk_buffer + CHUNK_MAX)) {
                        log(0, "chunk buffer overflow\n");
                        byte = 0xb4;  // sequence_error_code
                        break;
                        }
                  shift = (shift | byte) << 8;
                  if (current == end)
                        return;
                  }
            parseChunk();
            code       = byte;
            chunk_ptr  = chunk_buffer;
            shift      = 0xffffff00;
            }
      }

//---------------------------------------------------------
//   init
//---------------------------------------------------------

void AudioStream::init()
      {
      audioSync = false;
      audioHeaderFound = false;
      header = 0;
      }

//---------------------------------------------------------
//   head_check
//    check for valid audio header
//---------------------------------------------------------

bool AudioStream::headCheck() const
      {
      if( (header & 0xffe00000) != 0xffe00000)
            return false;
      if(!((header>>17)&3))
            return false;
      if( ((header>>12)&0xf) == 0xf)
            return false;
      if( ((header>>10)&0x3) == 0x3 )
            return false;
      return true;
      }

//---------------------------------------------------------
//   decodeHeader
//    decode audio header
//    returns segment len
//    return -1 on error
//---------------------------------------------------------

static int tabsel_123[2][3][16] = {
   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
     {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
     {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
      };
static long freqs[9] = {
      44100, 48000, 32000,
      22050, 24000, 16000,
      11025, 12000, 8000
      };

int AudioStream::decodeHeader()
      {
      bool mpeg25;
      int sampling_frequency;
      int error_protection;
      int bitrate_index;

      if (header & (1<<20) ) {
            lsf    = (header & (1<<19)) ? 0 : 1;
            mpeg25 = false;
            }
      else {
            lsf = 1;
            mpeg25 = true;
            }

      layer = 4 - ((header >> 17) & 3);
      if (((header>>10)&0x3) == 0x3) {
            log(1, "MpegAudio: Stream error\n");
            return -1;
            }
      if (mpeg25) {
            sampling_frequency = 6 + ((header>>10)&0x3);
            }
      else {
            sampling_frequency = ((header>>10)&0x3) + (lsf*3);
            }

      samplerate       = freqs[sampling_frequency];
      error_protection = ((header>>16)&0x1)^0x1;

      if (mpeg25)       /* allow Bitrate change for 2.5 ... */
            bitrate_index = ((header>>12)&0xf);

      bitrate_index = ((header>>12)&0xf);
      int padding   = ((header>>9)&0x1);
      bitrate       = tabsel_123[lsf][layer-1][bitrate_index];

      if (!bitrate_index) {
            log(1,"AudioHeader: Free format not supported.\n");
            return -1;
            }

      int frameSize;

      switch(layer) {
            case 1:
                  frameSize  = bitrate * 12000;
                  frameSize /= samplerate;
                  frameSize  = ((frameSize+padding)<<2)-4;
                  break;
            case 2:
                  frameSize = bitrate * 144000;
                  frameSize /= samplerate;
                  frameSize += padding - 4;
                  break;
            case 3:
                  frameSize  = bitrate * 144000;
                  frameSize /= samplerate << (lsf);
                  frameSize += padding - 4;
                  break;
            default:
                  log(1,"Sorry, unknown audio layer type.\n");
                  return -1;
            }
//      fprintf(stderr, "Audio Header layer %d  samplerate %d  framesize %d bitrate %dKBit/sec\n",
//         layer, samplerate, frameSize, bitrate);
      audioHeaderFound = true;
      return frameSize;
      }

//---------------------------------------------------------
//   putData
//---------------------------------------------------------

void AudioStream::putData(int len, FILE* ao)
      {
      static int hl = 0;
      static int dl = 0;

      while (len--) {
            int c = f.get();
            if (c == EOF) {
                  return;
                  }

            //-----------------------------------
            //  in sync mode output segBytesLeft
            //  to audio file
            //-----------------------------------

            if (audioSync) {
                  if (ao && apts)
                        putc(c, ao);
                  --segBytesLeft;
                  ++dl;
                  if (segBytesLeft == 0) {
                        if (state == FLUSH) {
                              state = FLUSH1;
                              return;
                              }
                        audioSync = false;
                        header    = 0;
                        dl = 0;
                        }
                  continue;
                  }

            //-----------------------------------
            //  search for new header
            //-----------------------------------

            ++hl;
            header = (header << 8) | (c & 0xff);
            if (!headCheck())
                  continue;

            //-----------------------------------
            //  header found
            //-----------------------------------

            if (hl > 4)
                  log(0, "   audio: lost %d bytes\n", hl - 4);
            hl = 0;
            segBytesLeft = decodeHeader();
            if (segBytesLeft == -1) { // bad audio header
                  log(1,"Bad audio header.\n");
                  continue;
                  }

            // output header:
            if (ao && apts) {
                  putc(header >> 24, ao);
                  putc(header >> 16, ao);
                  putc(header >> 8, ao);
                  putc(header, ao);
                  }

            audioSync = true;
            audioHeaderFound = true;

            // 1152 samples block
            if (apts) {
                  int64_t t = 1152 * 90000 / samplerate;
                  apts  += t;
                  adpts += t;
                  }
            }
      }

//---------------------------------------------------------
//   encode
//---------------------------------------------------------

static void encode(FILE* ao)
      {
      chunk_ptr = chunk_buffer;
      code      = 0xb4;
      shift     = 0;
      in_slice  = false;
      is_sequence_needed = true;

      startTime = time(0);
      doubledPictures = 0;
      droppedPictures = 0;

      while (nextStartCode()) {
            uint64_t pts = 0;
            unsigned streamId = f.get();
            int len           = ((f.get() & 0xff) << 8) + (f.get() & 0xff);
            f.get();
            int flags2        = f.get();
            int headerLen     = f.get();
            len -= (headerLen + 3);
            int n = 0;
            if (flags2 & 0xc0) {
                  uint64_t a = f.get();
                  pts = (a & 0xe0) << 29;
                  a = ((f.get() & 0xff) << 8) + (f.get() & 0xff);
                  pts |= (a & 0xfffe) << 14;
                  a = ((f.get() & 0xff) << 8) + (f.get() & 0xff);
                  pts |= (a & 0xfffe) >> 1;
                  n += 5;
                  if ((flags2 & 0xc0) == 0xc0) {
                        f.get();    // skip dts
                        f.get();
                        f.get();
                        f.get();
                        f.get();
                        n += 5;
                        }
                  }
            if (f.eof())
                  break;
            for (int i = n; i < headerLen; ++i)
                  f.get();
            if (f.eof())
                  break;
            if (flags2 & 0x3f) {
                  log(0, "bad PES packet, flags2=%02x, skip %d bytes\n",
                          flags2, len);
                  continue;   // search for next start code
                  }

            if ((streamId & 0xf0) == 0xe0) {
                  if (pts) {
                        vpts = pts;
                        }
                  videoSequence(len);
                  }
            else if ((streamId & 0xe0) == 0xc0) {
                  if (streamId != 0xc0) {
                        while(len--)
                              f.get();
                        }
                  else {
                        if (pts) {
                              if (apts != pts) {
                                    if (apts) {
                                          int64_t as = pts - apts;
                                          int64_t asa = as > 0 ? as : - as;
                                          if (asa > 10)
                                                log(1, "audio stream break %lld = %lld - %lld (%d pics)\n",
                                                   as, pts, apts, int(as/3600));
                                          async += (pts - apts);
                                          }
                                    apts = pts;
                                    }
                              }
                        audio.putData(len, ao);
                        }
                  }
            else if (streamId == 0xbd) {   // ??
                  while(len--)
                        f.get();
                  }
            else {
                  log(0, "unknown stream id 0x%02x\n", streamId);
                  while(len--)
                        f.get();
                  }
            if ((state == INIT) && audio.audioHeaderFound && picture_header_found)
                  return;
            if (state == FLUSH1)
                  return;
            if (state == RESTART) {
                  log(0, "encode: RESTART\n");
                  return;
                  }
            }
      log(0, "STOP: EOF found!\n");
      }

//---------------------------------------------------------
//   videoloop
//    video thread
//---------------------------------------------------------

static void* videoloop(void* mops)
      {
      pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
      pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);

      int f = open(pipename, O_RDONLY);
      if (f == -1) {
            fprintf(stderr, "\ncannot open pipe <%s>: %s\n",
               pipename, strerror(errno));
            pthread_exit((void*)-1);
            }
      char voname[512];
      strcpy(voname, oName);
      strcat(voname, ".mpv");
      int out = open(voname, O_WRONLY | O_CREAT, 0644);
      if (out == -1) {
            fprintf(stderr, "\ncannot open <%s>: %s\n",
               voname, strerror(errno));
            pthread_exit((void*)-1);
            }
      const int bs = 4096;
      videoWritten = 0;
      char buffer[bs];

      int error = 0;
      for (;;) {
            int n = read(f, buffer, bs);
            if (n <= 0) {
                  if (n == 0) {
                        log(1, "video loop: read returns 0: %s\n",
                                 strerror(errno));
                        if (errno == EAGAIN)
                              continue;
                        break;
                        }
                  else {
                        log(1, "\nvideoloop: read error: %s\n", strerror(errno));
                        error = 1;
                        break;
                        }
                  }
            int rv = write(out, buffer, n);
            if (rv != n) {
                  log(1, "\noutput return %d write error: %s\n",
                           rv, strerror(errno));
                  error = 1;
                  break;
                  }
            videoWritten += rv;
            }
      close(out);
      close(f);
      log(0, "exit video loop return code %d\n", error);
      pthread_exit((void*) error);
      return 0;
      }

//---------------------------------------------------------
//   printVersion
//---------------------------------------------------------

static void printVersion(const char* prog)
      {
      fprintf(stderr, "%s: Version %s\n", prog, VERSION);
      }

//---------------------------------------------------------
//   usage
//---------------------------------------------------------

static void usage(const char* prog, const char* txt)
      {
      fprintf(stderr, "%s: %s\nusage: %s flags record-directory\n  Flags:\n",
         prog, txt, prog);
      fprintf(stderr, "   -V        print version\n");
      fprintf(stderr, "   -v        verbose output\n");
      fprintf(stderr, "   -M        generate mpegs only\n");
      fprintf(stderr, "   -1        generate vcd instead of svcd\n");
      fprintf(stderr, "   -o name   output file name\n");
      fprintf(stderr, "   -q nn     start quality setting for encoder default: 4\n");
      fprintf(stderr, "   -s nn     CD-Size (MB) default: 800\n");
      fprintf(stderr, "   -m        monochrome input\n");
      fprintf(stderr, "   -w nn     widescreen (blank nn*4 lines from top)\n");
      fprintf(stderr, "   -n        denoise\n");
      fprintf(stderr, "   -o name   mpeg video output file\n");
      fprintf(stderr, "   -c nn     number of output CD's\n");
      fprintf(stderr, "   -U        ignore marks.vdr; transcode whole recording\n");
      fprintf(stderr, "   -k        dont't delete temp files\n");
      fprintf(stderr, "   -r        dont restart if output does not fit on cds\n");
      fprintf(stderr, "   -a        transcode audio to 44100 sample rate\n");
      }

//---------------------------------------------------------
//   buildPipeCmd
//---------------------------------------------------------

void buildPipeCmd()
      {
      //---------------------------------------------------
      //  prepare pipe to encoder
      //    | yuvdenoise | yuvscaler | mpeg2enc
      //---------------------------------------------------

      pipeCmd[0] = 0;
      char* p = pipeCmd;

      if (denoise) {
            if (monochrome)
                  strcat(p, "yuvdenoise -B | ");
            else
                  strcat(p, "yuvdenoise | ");
            while (*p)
                  ++p;
            }

      if (mpeg2dec->width() != twidth || mpeg2dec->height() != theight || widescreen) {
            strcat(p, "yuvscaler -v0");
            while(*p)
                  ++p;
            if (widescreen) {
                  if (mpeg2dec->width() == twidth && mpeg2dec->height() == theight) {
                        sprintf(p, " -I ACTIVE_%dx%d+0+%d -M RATIO_1_1_1_1 |",
                           mpeg2dec->width(), theight - widescreen*2, widescreen);
                        }
                  else {
                        if (vcdonly)
                              sprintf(p, " -I ACTIVE_%dx%d+0+%d -O VCD -M MMX |",
                                 mpeg2dec->width(), theight - widescreen*2, widescreen);
                        else
                              sprintf(p, " -I ACTIVE_%dx%d+0+%d -O SVCD -M MMX |",
                                 mpeg2dec->width(), theight - widescreen*2, widescreen);
                        }
                  }
            else
                  sprintf(p, " -O %sVCD -M MMX |", vcdonly ? "" : "S");
            }
      while (*p)
            ++p;
      //
      // max Bitrate for SVCD: 2718 kBit/sec:
      //
      //
      int ar = wideScreen ? 3 : mpeg2dec->aspectRatio();
      if (vcdonly) {
            if (denoise2)
                  sprintf(p, "mpeg2enc -N -a %d -S %d -B %d -I0 -q %d -v0 -f 1 %s",
                     ar, cdsize, audio.bitrate, quality,
		     mpegonly ? "-d" : "");
            else
                  sprintf(p, "mpeg2enc -a %d -S %d -B %d -I0 -q %d -v0 -f 1 %s",
                     ar, cdsize, audio.bitrate, quality,
		     mpegonly ? "-d" : "");
            }
      else {
            int videoBitrate = 2718 - audio.bitrate;
            if (denoise2)
                  sprintf(p, "mpeg2enc -N -a %d -S %d -B %d -I0 -q %d -v0 -f 4 -b %d %s",
                     ar, cdsize, audio.bitrate, quality,
                     videoBitrate, mpegonly ? "-d" : "");
            else
                  sprintf(p, "mpeg2enc -a %d -S %d -B %d -I0 -q %d -v0 -f 4 -b %d %s",
                     ar, cdsize, audio.bitrate, quality,
                     videoBitrate, mpegonly ? "-d" : "");
            }
      while (*p)
            ++p;
      sprintf(p, " -o %s", pipename);
      }

//---------------------------------------------------------
//   encodeLoop
//---------------------------------------------------------

void encodeLoop()
      {
      char aoname[512];
      strcpy(aoname, oName);
      strcat(aoname, ".mpa");

      for (;;) {        // loop over retries with different q-values
            curPic = 0;

            audio.init();
            FILE* ao = fopen(aoname, "w");
            if (ao == 0) {
                  log(1, "cannot open audio output <%s>: %s\n",
                     aoname, strerror(errno));
                  }
            buildPipeCmd();
            log(0, "popen <%s>\n", pipeCmd);
            vPipe = popen(pipeCmd, "w");
            if (vPipe == 0) {
                  log(0, "\npopen failed: %s", strerror(errno));
                  exit(-1);
                  }
            pthread_create(&videothread, 0, ::videoloop, 0);

            int z     = 0;
            int n     = 1;
            int arz, arn;
            mpeg2dec->aspect(arz, arn);
            mpeg2dec->frameRate(z, n);
            int width  = mpeg2dec->width();
            int height = mpeg2dec->height();
            bool prog  = mpeg2dec->progressiveFrame();

            fprintf(vPipe, "YUV4MPEG2 W%d H%d A%d:%d F%d:%d I%c\n",
               width, height, arz, arn, z, n, prog ? 'p' : 't');

            adpts = 0;   // audio device output state
            vdpts = 0;   // video device output state

            for (iMark curMark = markList.begin(); curMark != markList.end();) {
                  int startPicture = *curMark++;
                  endPicture       = *curMark++;

                  f.seekPic(startPicture);

                  absPic = startPicture;
                  state  = SYNC;
                  apts   = 0;
                  vpts   = 0;
                  vptse   = 0;
                  audio.init();
                  encode(ao);
                  if (state == RESTART)
                        break;
                  }
            if (state == RESTART)
                  log(0, "encodeLoop: break, restart\n");

            if (fclose(ao) == -1) {
                  perror("\nclose audio");
                  log(0, "close audio failed: %s\n", strerror(errno));
                  }
            int rv = pclose(vPipe);
            if (rv == -1) {
                  perror("\npclose video pipe");
                  log(0, "close pipe failed: %s\n", strerror(errno));
                  }
            log(0, "thread join\n");
            if (pthread_join(videothread, 0))
                  perror("Failed to join video thread");
            log(0, "nach thread join\n");

            if (state != RESTART) {
                  long long curSec = curPic/25;
                  int percent = 0;
                  if (curSec) {
                        long long cursize = (videoWritten/1024) + curSec * (audio.bitrate/8);
                        long long tsize   = cursize * (long long)(totalPictures/25) / curSec;
                        percent           = (tsize * 100LL)/ (cdsize*1024);
                        }

                  int tooBig = percent - 100*numcds;
                  if (autorestart && tooBig > 0) {
                        quality++;
                        state = RESTART;
                        }
                  }

            if (state != RESTART)
                  break;
            }
      }

//---------------------------------------------------------
//   main
//---------------------------------------------------------

int main(int argc, char* argv[])
      {
      int c;

      while ((c = getopt(argc, argv, "VvMo:q:s:mw:nN1Uc:kWraS:")) != EOF) {
            switch (c) {
                  case 'V': printVersion(argv[0]); exit(0);
                  case 'v': verbose    = true; break;
		      case 'M': mpegonly   = true; break;
                  case 'o': oName      = optarg; break;
                  case 'q': quality    = atoi(optarg); break;
                  case 's': cdsize     = atoi(optarg); break;
                  case 'm': monochrome = true; break;
                  case 'w': widescreen = atoi(optarg)*4; break;
                  case 'n': denoise    = true; break;
                  case 'N': denoise2   = true; break;
                  case '1': vcdonly    = true; break;
                  case 'U': whole      = true; break;
                  case 'c': numcds     = atoi(optarg); break;
                  case 'k': keepFiles  = true; break;
                  case 'W': wideScreen = true; break;
                  case 'r': autorestart = false; break;
                  case 'a': transcodeAudio = true; break;

                  default:
                        usage(argv[0], "bad argument");
                        exit(-1);
                  }
            }
      if (argc < 2) {
            usage(argv[0], "bad argument count");
            exit(-1);
            }
      argc -= optind;
      char* arg1 = *(argv+optind);

      if (vcdonly) {
            twidth  = 352;
            theight = 288;
            }
      char logname[512];
      sprintf(logname, "%s.log", oName);
      logf = fopen(logname, "w");
      time_t t = time(0);
      log(0, "tosvcd %s: %s", VERSION, ctime(&t));
      log(0, "encoding dir:  %s\n", arg1);
      log(0, "encoding name: %s\n", oName);

      //---------------------------------------------------
      //  read list of 0??.vdr files
      //---------------------------------------------------

      f.curFile = 0;
      struct stat fileStat;
      int rv = stat(arg1, &fileStat);
      if (rv == -1) {
            log(1, "cannot stat <%s>: %s\n", arg1, strerror(errno));
            exit(-1);
            }
      if (!S_ISDIR(fileStat.st_mode)) {
            log(1, "<%s> is not a directory\n", arg1);
            exit(-1);
            }

      glob_t flist;
      char* buf = new char[strlen(arg1)+11];
      sprintf(buf, "%s/0??.vdr", arg1);
      glob(buf, 0, 0, &flist);
      f.files = flist.gl_pathc;
      f.filenames = flist.gl_pathv;

      if (f.files == 0) {
            log(1, "no recordings 0??.vdr found in dir <%s>\n", arg1);
            exit(-1);
            }

      //---------------------------------------------
      // read index.vdr
      //---------------------------------------------

      sprintf(buf, "%s/index.vdr", arg1);
      int idx = open(buf, O_RDONLY);
      if (idx == -1) {
            log(1, "cannot open index file <%s>: %s\n",
               buf, strerror(errno));
            exit(-1);
            }
      rv = fstat(idx, &fileStat);
      totalPictures = fileStat.st_size / sizeof(Index);
      picIndex = new Index[totalPictures];
      rv = read(idx, picIndex, sizeof(Index) * totalPictures);
      if ((unsigned)rv != sizeof(Index) * totalPictures) {
            log(1, "error reading index <%s>: %s\n",
               buf, strerror(errno));
            exit(-1);
            }
      close(idx);
#if 0
      // debug:
      for (int i = 11250; i < 11500; ++i) {
            if (picIndex[i].type == I_FRAME) {
                  int h, m, s, f;
                  s = i / 25;
                  f = i % 25;
                  m = s / 60;
                  s = s % 60;
                  h = m / 60;
                  m = m % 60;
                  printf("%d:%02d:%02d.%02d\n", h, m, s, f);
                  break;
                  }
            }
      for (int i = 11500; i < 11700; ++i) {
            if (picIndex[i].type == I_FRAME) {
                  int h, m, s, f;
                  s = i / 25;
                  f = i % 25;
                  m = s / 60;
                  s = s % 60;
                  h = m / 60;
                  m = m % 60;
                  printf("%d:%02d:%02d.%02d\n", h, m, s, f);
                  break;
                  }
            }
#endif
      //---------------------------------------------
      //  read marks.vdr
      //---------------------------------------------

      if (!whole) {
            sprintf(buf, "%s/marks.vdr", arg1);
            FILE* mf = fopen(buf, "r");
            if (mf) {
                  char line[64];
                  while(fgets(line, 63, mf)) {
                        if (*line == 0 || *line == '\n')
                              continue;
                        int h, m, s, f;
                        int n = sscanf(line, "%d:%2d:%2d.%2d", &h, &m, &s, &f);
                        if (n != 4) {
                              log(1, "misformed marks file <%s> <%s>\n", buf, line);
                              exit(-1);
                              }
                        int pic = h * 3600 * 25 + m * 60 * 25 + s * 25 + f - 1;
                        markList.push_back(pic);
                        }
                  fclose(mf);
                  int len = markList.size();
                  if (len % 2)      // append missing end mark
                        markList.push_back(totalPictures);
                  }
            else {
                  if (verbose)
                        log(1, "no marks.vdr found, (re)coding whole movie\n");
                  }
            }
      if (markList.empty()) {
            markList.push_back(0);
            markList.push_back(totalPictures);
            }
      delete buf;

      //---------------------------------------------------
      //  compute real (cutted) len
      //---------------------------------------------------

      totalPictures = 0;
      for (iMark i = markList.begin(); i != markList.end();) {
            int start = *i++;
            int end   = *i++;
            totalPictures += end - start;
            }
      log(1, "total time: %d:%02d\n",
         totalPictures/(25*60),
         (totalPictures%(25*60))/25);

      if (f.open()) {
            log(1, "cannot open source\n");
            exit(-1);
            }

      if (numcds == -1) {
            // determine number of output CD's assuming
            // 60 min /CD    25 frames/sec
            int totalMinutes = totalPictures / (25*60);
            numcds = (totalMinutes+minutesPerCD-1) / minutesPerCD;
            if (numcds == 0)
                  numcds = 1;
            }

      //---------------------------------------------------
      //  init mpeg decoder
      //---------------------------------------------------

      int accel =  mm_accel();
      log(0, "Accel flags 0x%04x\n", accel);
      mpeg2dec = new MpegDec(accel);

      //---------------------------------------------------
      //  parse input stream to find out:
      //    - initial audio/video offset
      //    - video & audio properties
      //---------------------------------------------------

      state = INIT;
      audio.init();

      iMark curMark = markList.begin();
      int startPicture = *curMark++;
      f.seekPic(startPicture);

      encode(0);

      //---------------------------------------------------
      //    create pipes
      //---------------------------------------------------

      pipename = tempnam(".", "pipe");
      rv = mknod(pipename, S_IFIFO | 0600, 0);
      if ((rv == -1) && (errno != EEXIST)) {
            log(1, "cannot create pipe <%s>: %d %s\n", pipename, errno, strerror(errno));
            exit(-1);
            }

      //---------------------------------------------------
      //  encode
      //    process input stream
      //---------------------------------------------------

      int z = 0;
      int n = 1;
      int arz, arn;
      mpeg2dec->aspect(arz, arn);
      mpeg2dec->frameRate(z, n);
      bool prog  = mpeg2dec->progressiveFrame();
      int width  = mpeg2dec->width();
      int height = mpeg2dec->height();

      const char* ll = "===================================================================\n";
      log(1, ll);
      log(1, "Video: %dx%d  aspect: %d:%d  frame_rate: %d:%d  scan: %s\n",
               width, height, arz, arn, z, n, prog ? "progressive" : "interlaced");
      log(1, "Audio: layer %d  %d Samples/sec  %d KBit/sec\n",
               audio.layer, audio.samplerate, audio.bitrate);
      log(1, ll);

      encodeLoop();

      f.close();  // close input
      log(1, "\nfinished: dropped %d doubled %d Pictures\n",
         droppedPictures, doubledPictures);

      remove(pipename);     // ignore errors

      //---------------------------------------------------
      //  transcode audio
      //---------------------------------------------------

      if (transcodeAudio) {
            char ta[512];
            sprintf(ta, "mpg123 -w - '%s.mpa' | sox -c 2 -s -r 48000 -w -t raw - -r 44100"
              " -t wav - | toolame -e -p 2 -b 224 - '%s.mpa.44100'", oName, oName);
            if (system(ta)) {
                  log(1, "transcode audio failed\n");
                  log(1, "  cmd was: %s\n", ta);
                  return -1;
                  }
            }

      //---------------------------------------------------
      //  mplex
      //---------------------------------------------------

      char mplex[512];
      if (transcodeAudio)
            sprintf(mplex, "mplex -v0 -f%d -o '%s-%%d.mpg' '%s.mpa.44100' '%s.mpv'",
               vcdonly ? 1 : 4, oName, oName, oName);
      else
            sprintf(mplex, "mplex -v0 -f%d -o '%s-%%d.mpg' '%s.mpa' '%s.mpv'",
               vcdonly ? 1 : 4, oName, oName, oName);
      if (system(mplex)) {
            log(1, "mplex failed: <%s>\n", mplex);
            return -1;
            }

      if (!keepFiles) {
            // remove temp files:
            sprintf(mplex, "%s.mpv", oName);
            remove(mplex);

            sprintf(mplex, "%s.mpa", oName);
            remove(mplex);

            if (transcodeAudio) {
                  sprintf(mplex, "%s.mpv.44100", oName);
                  remove(mplex);
                  }
            }

      //---------------------------------------------------
      //  vcdimager
      //---------------------------------------------------

      glob_t mflist;
      char* vbuf = new char[strlen(oName)+7];
      sprintf(vbuf, "%s-*.mpg", oName);
      glob(vbuf, 0, 0, &mflist);
      n = mflist.gl_pathc;
      if (n == 0) {
            log(0, "no generated mpeg files found!\n");
            fclose(logf);
            return 0;
            }
      if (!mpegonly) {
	    for (int i = 0; i < n; ++i) {
		  char* name = mflist.gl_pathv[i];
		  char buffer[512];
		  sprintf(buffer, "vcdimager --type=%s --quiet --volume-count=%d "
		     "--volume-number=%d --cue-file='%s-%d.cue' "
		     "--bin-file='%s-%d.bin' '%s'",
		     vcdonly ? "vcd2" : "svcd",
		     n, i+1, oName, i+1, oName, i+1, name);
		  if (system(buffer)) {
		      log(1, "vcdimager failed: <%s>: %s\n",
                     buffer, strerror(errno));
                  fclose(logf);
                  return 0;
			}
		  if (!keepFiles)
			remove(name);
		  }
	      }
      if (logf) {
            log(0, "finished successfully\n");
            fclose(logf);
            }
      return 0;
      }

